@@ -1314,10 +1314,7 @@ do_decode(PyObject* decoder, const char* jsonStr, Py_ssize_t jsonStrLen,
         reader.Parse<kParseInsituFlag | kParseNumbersAsStringsFlag>(ss, handler);
 
     if (reader.HasParseError()) {
-        SizeType offset = reader.GetErrorOffset();
-        ParseErrorCode code = reader.GetParseErrorCode();
-        const char* msg = GetParseError_En(code);
-        const char* fmt = "Parse error at offset %d: %s";
+        size_t offset = reader.GetErrorOffset();
 
         if (PyErr_Occurred()) {
             PyObject* etype;
@@ -1328,16 +1325,18 @@ do_decode(PyObject* decoder, const char* jsonStr, Py_ssize_t jsonStrLen,
             // Try to add the offset in the error message if the exception
             // value is a string.  Otherwise, use the original exception since
             // we can't be sure the exception type takes a single string.
-            const char* emsg = msg;
-            if (PyUnicode_Check(evalue)) {
-                emsg = PyUnicode_AsUTF8(evalue);
-                PyErr_Format(etype, fmt, offset, emsg);
+            if (evalue != NULL && PyUnicode_Check(evalue)) {
+                PyErr_Format(etype, "Parse error at offset %zu: %S", offset, evalue);
+                Py_DECREF(etype);
+                Py_DECREF(evalue);
+                Py_XDECREF(etraceback);
             }
             else
                 PyErr_Restore(etype, evalue, etraceback);
         }
         else
-            PyErr_Format(PyExc_ValueError, fmt, offset, msg);
+            PyErr_Format(PyExc_ValueError, "Parse error at offset %zu: %s",
+                         offset, GetParseError_En(reader.GetParseErrorCode()));
 
         Py_XDECREF(handler.root);
         free(jsonStrCopy);
@@ -1545,6 +1544,13 @@ dumps_internal(
 #define RECURSE(v) dumps_internal(writer, v, skipKeys, defaultFn, sortKeys, \
                                   numberMode, datetimeMode, uuidMode)
 
+#define ASSERT_VALID_SIZE(l) do {                                       \
+    if (l < 0 || l > UINT_MAX) {                                        \
+        PyErr_SetString(PyExc_ValueError, "Out of range string size");  \
+        return false;                                                   \
+    } } while(0)
+
+
     if (object == Py_None) {
         writer->Null();
     }
@@ -1664,15 +1670,16 @@ dumps_internal(
 
         if (PyBytes_AsStringAndSize(object, &s, &l) == -1)
             return false;
-
-        writer->String(s, l);
+        ASSERT_VALID_SIZE(l);
+        writer->String(s, (SizeType) l);
     }
     else if (PyUnicode_Check(object)) {
         Py_ssize_t l;
         const char* s = PyUnicode_AsUTF8AndSize(object, &l);
         if (s == NULL)
             return false;
-        writer->String(s, l);
+        ASSERT_VALID_SIZE(l);
+        writer->String(s, (SizeType) l);
     }
     else if (PyList_Check(object)) {
         writer->StartArray();
@@ -1722,7 +1729,8 @@ dumps_internal(
                     const char* key_str = PyUnicode_AsUTF8AndSize(key, &l);
                     if (key_str == NULL)
                         return false;
-                    writer->Key(key_str, l);
+                    ASSERT_VALID_SIZE(l);
+                    writer->Key(key_str, (SizeType) l);
                     if (Py_EnterRecursiveCall(" while JSONifying dict object"))
                         return false;
                     bool r = RECURSE(item);
@@ -1745,6 +1753,7 @@ dumps_internal(
                     const char* key_str = PyUnicode_AsUTF8AndSize(key, &l);
                     if (key_str == NULL)
                         return false;
+                    ASSERT_VALID_SIZE(l);
                     items.push_back(DictItem(key_str, l, item));
                 }
                 else if (!skipKeys) {
@@ -1755,8 +1764,8 @@ dumps_internal(
 
             std::sort(items.begin(), items.end());
 
-            for (int i=0, s=items.size(); i < s; i++) {
-                writer->Key(items[i].key_str, items[i].key_size);
+            for (size_t i=0, s=items.size(); i < s; i++) {
+                writer->Key(items[i].key_str, (SizeType) items[i].key_size);
                 if (Py_EnterRecursiveCall(" while JSONifying dict object"))
                     return false;
                 bool r = RECURSE(items[i].item);
@@ -2031,7 +2040,7 @@ dumps_internal(
             Py_DECREF(retval);
             return false;
         }
-        writer->String(s, l);
+        writer->String(s, (SizeType) l);
         Py_DECREF(retval);
     }
     else if (PyIter_Check(object)) {
@@ -2087,6 +2096,7 @@ dumps_internal(
     return true;
 
 #undef RECURSE
+#undef ASSERT_VALID_SIZE
 }
 
 
